from typing import Optional
from fastapi import Query, Depends


class CommonQueryPagination:
    def __init__(
        self,
        page: int = Query(default=1, ge=1),
        limit: int = Query(default=20, gt=0, le=100),
    ):

        self.page = page
        self.limit = limit


class CommonQuerySelectFields:
    def __init__(self, fields: Optional[str] = Query(default=None)):
        self.fields = fields


class CommonQueryFilter:
    def __init__(
        self, sortBy: str = Query(default="id"), sortType: str = Query(default="asc")
    ):
        self.sortBy = sortBy
        self.sortType = sortType


class CommonQueryGetter:
    def __init__(
        self,
        fields: Optional[str] = Query(default=None),
        page: int = Query(default=1, ge=1),
        limit: int = Query(default=20, gt=0, le=100),
        sortBy: str = Query(default="id"),
        sortType: str = Query(default="asc"),
    ):
        self.fields = fields
        self.page = page
        self.limit = limit
        self.sortBy = sortBy
        self.sortType = sortType



from fastapi import FastAPI, APIRouter, Depends, Request, Path
from fastapi.encoders import SetIntStr, DictIntStrAny
from fastapi.routing import APIRoute
from fastapi.datastructures import Default
from pydantic import BaseConfig, BaseModel, create_model
from pydantic.fields import ModelField
from starlette.routing import BaseRoute
from starlette.responses import Response, JSONResponse
from starlette.types import ASGIApp
from typing import Optional, List, Sequence, Type, Any, Callable, Union, Dict
from sqlalchemy.orm import decl_api
from sqlalchemy.ext.asyncio import AsyncSession
from types import FunctionType
from datetime import datetime

from .dependencies.utils import BaseCRUD
from .dependencies.utility import CommonQueryGetter

from .routing import *


class RouterMap():
    """
    Router Mapping Class to define and simplify your router simple CRUD
    """
    __updated_routers = {}

    @classmethod
    def _get_variable(cls):
        return [cls.__dict__[v] for v in vars(cls) if type(cls.__dict__[v]) in RouterClasses]
    
    @classmethod
    def _get_key_value(cls):
        return {v: cls.__dict__[v] for v in vars(cls) if type(cls.__dict__[v]) in RouterClasses}

    @classmethod
    def _collect_simple_router(cls):
        allRouters = {}
        for c in cls.__subclasses__():
            for v in c._get_key_value().values():
                allRouters[v.tablename] = v
            if c.__subclasses__():
                allRouters.update(c._collect_simple_router())
        if cls == RouterMap:
            allRouters.update(RouterMap.__updated_routers)
        return allRouters
    
    @staticmethod
    def create_router_map_from_base(
            base: decl_api.DeclarativeMeta,
            base_prefix: str = "",
            extend: bool = False
        ):
        """
        Create All CRUD Automatically from SQLAlchemy declared base
        
        :params:
        - base -> SQLAlchemy declarative base
        - base_prefix -> Base prefix path for all endpoints in one router
        - extend -> Applying ExtendedRouter for more API
        """
        class AutoMap(RouterMap): pass
        targetRouter = ExtendedRouter if extend else SimpleRouter 
        for c in base.__subclasses__():
            setattr(
                AutoMap,
                c.__tablename__,                
                targetRouter(c, prefix=base_prefix+"/"+c.__tablename__, tags=[c.__tablename__])
                )
        return AutoMap
    
    @classmethod
    def update_map(cls, simple_router: SimpleRouterType):
        """
        Add your fastapi_simple_crud.SimpleRouter object to be generated outside of the router map
        """
        if cls == RouterMap:
            RouterMap.__updated_routers[simple_router.tablename] = simple_router
        else:
            setattr(cls, simple_router.tablename, simple_router)

    @classmethod
    def generate(
            cls,
            application: Optional[FastAPI] = None,
            session_getter: Optional[FunctionType] = None
        ):
        """
        Generate routers that have been defined using RouterMap
        
        :params
        - application -> FastAPI Application
        - session_getter -> SQLAlchemy AsyncSession Getter/yielder
        """
        if cls == RouterMap:
            return SimpleCRUDGenerator(application, session_getter, True)
        e = "RouterMap.generate() only able to be called from 'RouterMap' class"
        raise BaseException(e)


class SimpleCRUDGenerator():
    """
    Create CRUD Operation in a simple way.
    
    params:
    - application -> FastAPI Application
    - session_getter -> method to get the sqlalchemy session
    - autogenerate (bool) -> once instantiated, the base crud endpoints will be created
    """
    def __init__(
            self,
            application: Optional[FastAPI] = None,
            session_getter: Optional[FunctionType] = None, 
            autogenerate: bool = True
        ):
        self.app = application
        self.allRouters = RouterMap._collect_simple_router()
        self.session_getter = session_getter
        if not all([application, session_getter]): autogenerate = False
        if autogenerate: self.generate_router()
    
    def set_application(self, application: FastAPI):
        """
        Set your FastAPI object application
        """
        self.app = application
    
    def set_session_getter(self, session_getter: FunctionType):
        """
        Set your session getter for SQLAlchemy
        """
        self.session_getter = session_getter

    def update_map(self, simple_router: SimpleRouterType):
        """
        Add your fastapi_simple_crud.SimpleRouter object to be generated outside of the router map
        """
        self.allRouters[simple_router.tablename] = simple_router

    def generate_router(
            self,
            application: Optional[FastAPI] = None,
            session_getter: Optional[FunctionType] = None
        ):
        """
        Generate defined router map
        """
        if application: self.set_application(application)
        if session_getter: self.set_session_getter(session_getter)
        if all([self.app, self.session_getter]):
            for tag in sorted(self.allRouters):
                router = self.allRouters[tag]
                router.set_the_get_session(self.session_getter)
                router._setup_crud()
                self.app.include_router(router)


import fastapi
from datetime import datetime
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import load_only, decl_api
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy import asc, desc, func, Column
from sqlalchemy.sql.selectable import Select
from typing import Any, List, Optional, Union, Dict
from pydantic import create_model

from .status import StatusResponse
from .utility import CommonQueryGetter
from .log import logger


def create_status(code, message_server, message_client=None):
    if message_client is None:
        message_client = message_server
    return {"code": code, "message": message_server}


def create_response(data={}, meta={}, status={}):
    return {
        "data": data,
        "meta": meta,
        "status": status,
    }

async def update_data(session, tableModel, dataObject, newDataPydantic):
    updates = {}
    fields = [i for i in vars(tableModel) if "_" not in [i[0], i[-1]]]
    for key in fields:
        if key in newDataPydantic.__dict__:
            updates[key] = newDataPydantic.__dict__[key]
    for key, value in updates.items():
        setattr(dataObject, key, value)
    session.add(dataObject)
    await session.commit()
    await session.refresh(dataObject)

class StatusCreator(StatusResponse):
    """
    to add more status, edit dependencies/status.py in StatusResponse class
    """

    def __init__(self):
        self._applyStatusAsAttribute()

    def _applyStatusAsAttribute(self):
        allAttr = [i for i in vars(StatusResponse) if not "_" in [i[0], i[-1]]]
        for a in allAttr:
            def fetcher(attrName):
                def response(message=str(attrName).replace("_", " ")):
                    code = StatusResponse.__dict__[attrName]
                    return create_status(code, str(message))
                return response
            self.__dict__[a] = fetcher(a)


status = StatusCreator()
"""
- The default message is the status name
- You can add your custom message by inserting your message to the function
    - ex: status.success("great")
- To add more status, edit 'StatusResponse' class in dependencies/status.py
"""


def get_field(classModel):
    try:
        fields = [i for i in vars(classModel) if "_" not in [i[0], i[-1]]]
        columnFields = []
        for f in fields:
            if "comparator" in classModel.__dict__[f].__dict__:
                if (
                    "column"
                    in str(classModel.__dict__[f].__dict__["comparator"]).lower()
                ):
                    columnFields.append(f)
        return columnFields
    except:
        return []

def validate_field(modelField, targetField):
    if type(modelField) in [list, tuple, dict]:
        availableField = modelField
    else:
        availableField = get_field(modelField)
    return [i for i in targetField if i in availableField]

class SimpleCRUDBaseDeclaration:
    base = None
    
    @staticmethod
    def set_base(base: decl_api.DeclarativeMeta):
        if type(base) == decl_api.DeclarativeMeta:
            SimpleCRUDBaseDeclaration.base = base
        else:
            raise BaseException("Invalid Base")

def set_declarative_base(base: decl_api.DeclarativeMeta):
    return SimpleCRUDBaseDeclaration.set_base(base)

class Selector:
    """
    Selector Class, modifiying the sqlalchemy select
    """

    def __init__(self, *columns, **columnsKeyPair):
        """
        add your customer key by adding the key-value pair
            ex: Selector(uid = User.id, name = Profile.name)

        or

        define it automatically by just giving the column name
            ex: Selector(User.id, Profile.name)

        (this output column name will be user_id, profile_name)
        """
        if not SimpleCRUDBaseDeclaration.base:
            msg = "SQLAlchemy Declarative Base is not defined. please define it by using 'set_declarative_base'\n"
            msg += "    >>> from fastapi_simple_crud.dependencies import utils\n"
            msg += "    >>> utils.set_declarative_base(Base)"
            raise BaseException(msg)
        self.baseClass = SimpleCRUDBaseDeclaration.base.__subclasses__()
        if columnsKeyPair:
            self.columns = list(columnsKeyPair.values())
            self.columnsKeyPair = columnsKeyPair
        else:
            self.columns = columns
            self.columnsKeyPair = dict(zip(self.get_keys(), list(columns)))
        self.keys = list(self.columnsKeyPair.keys())

    @property
    def query(self):
        return select(*self.columns)

    def get_keys(self):
        keys = []
        for c in self.columns:
            className, key = str(c).split(".")
            className = self.check_belonging(className)
            keys.append(className + "_" + key)
        return keys

    def check_belonging(self, key):
        for c in self.baseClass:
            cKey = str(c).split("'")[1].split(".")[-1]
            if key == cKey:
                return c.__tablename__
        return key

    async def execute(self, session: AsyncSession, query: Select):
        data = []
        query = await session.execute(query)
        for q in query:
            data.append(dict(zip(self.keys, list(q))))
        return data


class QueryManager:
    def __init__(self, classModel):
        self.classModel = classModel

    @property
    def fields(self) -> list:
        return get_field(self.classModel)

    @property
    def rawQuery(self) -> Select:
        return select(self.classModel)

    def validate_fields(self, fields: list) -> list:
        if fields:
            return [i for i in fields if i in self.fields]
        return []


def QueryPaginator(
    getParams: CommonQueryGetter, _obj: Union[decl_api.DeclarativeMeta, Selector]
):
    """
    Query Paginator generator for single or multiple table
    """
    if type(_obj) == Selector:
        return QueryPaginatorMultiple(getParams, _obj)
    if _obj.__class__ == decl_api.DeclarativeMeta:
        if len(_obj.__mro__) == 3:
            return QueryPaginatorSingle(getParams, _obj)
    raise BaseException("Error Paginator")


class QueryPaginatorSingle(QueryManager):
    """
    Query Paginator for common use of paginating functions
    """

    def __init__(self, getParams: CommonQueryGetter, classModel):
        QueryManager.__init__(self, classModel)
        self.getParams = getParams
        if getParams.fields:
            self.filterFields = self.validate_fields(getParams.fields.split(","))
        else:
            self.filterFields = self.fields

    def filter(self, query: Select, fields: Optional[list] = None) -> Select:
        """
        filter the desired fields from the query (it will remains the primary key)
        """
        if not fields:
            fields = self.filterFields
        return query.options(load_only(*fields))

    def filter_primary_key(self, data: list) -> list:
        """
        primary key filter
        """
        newData = []
        for d in data:
            newPair = {}
            for key in d.__dict__:
                if key in self.filterFields:
                    newPair[key] = d.__dict__[key]
            newData.append(newPair)
        return newData

    def sort(self, query: Select, sortBy: str, sortType: str) -> Select:
        """
        sort your query by defining both the sortBy 'fields' and sortType 'asc'/'desc' parameters
        """
        if sortBy in self.fields:
            merchantOrder = self.classModel.__dict__[sortBy]
            for iSortType, orderMethod in [["asc", asc], ["desc", desc]]:
                if sortType == iSortType:
                    return query.order_by(orderMethod(merchantOrder))
        return query

    def paginate(self, query: Select, pageNo: int, limitPerPage: int) -> Select:
        """
        paginate query by defining page number and page limit
        """
        query = query.offset((pageNo - 1) * limitPerPage)
        query = query.limit(limitPerPage)
        return query

    async def execute_pagination(self, session: AsyncSession, query: Select) -> dict:
        """
        sort, filter and paginate the query with executed query outputs
        """
        query = self.filter(query)
        countAfterFilter = await session.execute(
            select(func.count()).select_from(query.subquery())
        )
        countAfterFilter = countAfterFilter.scalars().one()
        query = self.sort(query, self.getParams.sortBy, self.getParams.sortType)
        query = self.paginate(query, self.getParams.page, self.getParams.limit)
        query = await session.execute(query)
        datas = query.scalars().all()
        datas = self.filter_primary_key(datas)
        return create_response(
            data={"list": datas},
            meta={
                "page": self.getParams.page,
                "length": self.getParams.limit,
                "total": countAfterFilter,
            },
            status=status.success(),
        )


class QueryPaginatorMultiple:
    def __init__(self, getParams: CommonQueryGetter, selector: Selector):
        self.rawQuery = selector.query
        self.getParams = getParams
        self.selector = selector
        if getParams.fields:
            self.fields = validate_field(
                self.selector.keys, getParams.fields.split(",")
            )
        else:
            self.fields = self.selector.keys

    def filter(self, data: list, fields: Optional[list] = None) -> Select:
        """
        filter the desired fields from the query (it will remains the primary key)
        """
        if not fields:
            fields = self.fields
        newData = []
        for d in data:
            newPair = {}
            for k in d:
                if k in fields:
                    newPair[k] = d[k]
            newData.append(newPair)
        return newData

    def sort(self, query: Select, sortBy: str, sortType: str) -> Select:
        """
        sort your query by defining both the sortBy 'fields' and sortType 'asc'/'desc' parameters
        """
        if sortBy in self.fields:
            merchantOrder = self.selector.columnsKeyPair[sortBy]
            for iSortType, orderMethod in [["asc", asc], ["desc", desc]]:
                if sortType == iSortType:
                    return query.order_by(orderMethod(merchantOrder))
        return query

    def paginate(self, query: Select, pageNo: int, limitPerPage: int) -> Select:
        """
        paginate query by defining page number and page limit
        """
        query = query.offset((pageNo - 1) * limitPerPage)
        query = query.limit(limitPerPage)
        return query

    async def execute_pagination(self, session: AsyncSession, query: Select):
        countAfterFilter = await session.execute(
            select(func.count()).select_from(query.subquery())
        )
        countAfterFilter = countAfterFilter.scalars().one()
        query = self.sort(query, self.getParams.sortBy, self.getParams.sortType)
        query = self.paginate(query, self.getParams.page, self.getParams.limit)
        datas = await self.selector.execute(session, query)
        datas = self.filter(datas)
        return create_response(
            data={"list": datas},
            meta={
                "page": self.getParams.page,
                "length": self.getParams.limit,
                "total": countAfterFilter,
            },
            status=status.success(),
        )

## Future update
# class BaseLikeClause:
#     def __init__(self, classModel, *likeExpression, **likeClause):
#         self.classModel = classModel
#         self.likeExp = likeExpression
#         self.lc = self.filterEmptyClause(likeClause)

#     def applyLikeObject(self, query: Select, whereClauseDict, **WhereClauseKwargs):
#         WhereClauseKwargs.update(whereClauseDict)
#         self.wc.update(WhereClauseKwargs)
#         for w in self.we:
#             query = query.filter(w)
#         for key in self.wc:
#             query = query.where(self.classModel.__dict__[key] == self.wc[key])
#         return query
    
#     def filterEmptyClause(self, likeClauseInDict: dict):
#         newDict = {}
#         for key, val in likeClauseInDict.items():
#             if val != None:
#                 newDict[key] = val
#         return newDict

class BaseWhereClause:
    def __init__(self, classModel, *whereExpression, **whereClause):
        self.classModel = classModel
        self.we = whereExpression
        self.wc = whereClause

    def applyWhereObject(self, query: Select, whereClauseDict: Optional[dict] = {}, **WhereClauseKwargs):
        WhereClauseKwargs.update(whereClauseDict)
        self.wc.update(WhereClauseKwargs)
        self.wc = self.filterEmptyClause(self.wc)
        for w in self.we:
            query = query.where(w)
        for key in self.wc:
            query = query.where(self.classModel.__dict__[key] == self.wc[key])
        return query
    
    def filterEmptyClause(self, whereClauseInDict: dict):
        newDict = {}
        for key, val in whereClauseInDict.items():
            if val != None:
                newDict[key] = val
        return newDict

class BaseCRUD:
    def __init__(self, classModel):
        self.classModel = classModel

    def where(self, *whereExpression, **whereClause):
        return BaseWhereClause(self.classModel, *whereExpression, **whereClause)

    async def read_one(
            self,
            pydanticModel,
            session: AsyncSession
        ):
        try:
            query = select(self.classModel)
            for k, attr in pydanticModel.dict().items():
                if k in vars(self.classModel):
                    query = query.where(vars(self.classModel)[k]==attr)
            data = await session.execute(query)
            data = data.scalars().first()
            res = create_response(data=data, status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res
    
    async def read_many(
            self,
            getParams: CommonQueryGetter,
            session: AsyncSession,
            whereClauseObject: Optional[BaseWhereClause] = None,
            **whereClause
        ):
        try:
            paginator = QueryPaginator(getParams, self.classModel)
            query = paginator.rawQuery
            if whereClauseObject:
                query = whereClauseObject.applyWhereObject(query, whereClause)
            res = await paginator.execute_pagination(session, query)
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def read(self, getParams: CommonQueryGetter, session: AsyncSession):
        """
        with pagination
        """
        try:
            paginator = QueryPaginator(getParams, self.classModel)
            query = paginator.rawQuery
            res = await paginator.execute_pagination(session, query)
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def create(self, pydanticModel, session: AsyncSession):
        try:
            newData = self.classModel(**pydanticModel.dict())
            session.add(newData)
            await session.commit()
            await session.refresh(newData)
            res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res
    
    async def create_many(self, pydanticModelCollection, session: AsyncSession):
        try:
            dataCollection = pydanticModelCollection.dict()[self.classModel.__tablename__]
            createdObj = []
            statuses = []
            successCreate = 0
            failCreate = 0
            for data in dataCollection:
                try:
                    obj = self.classModel(**data)
                    session.add(obj)
                    createdObj.append(obj)                    
                    successCreate += 1
                    statuses.append(status.success())
                    await session.flush()
                except Exception as e:
                    failCreate += 1
                    statuses.append(status.error(e))
            if not failCreate:
                await session.commit()
                for o in createdObj:
                    await session.refresh(o)
            res = create_response(
                data={"status":statuses},
                meta={
                    "succeed": successCreate,
                    "failed": failCreate
                },
                status=status.error() if failCreate else status.success()
                )
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def update(
            self,
            pydanticModel,
            id,
            session: AsyncSession,
            whereClauseObject: Optional[BaseWhereClause] = None,
            **whereClause
        ):
        try:
            query = select(self.classModel)
            if id!=None:
                query = query.where(self.classModel.id==id)
            if whereClauseObject:
                query = whereClauseObject.applyWhereObject(query, whereClause)
            data = await session.execute(query)
            data = data.scalars().first()
            if not data:
                res = create_response(status=status.data_is_not_updated())
            else:
                await update_data(session, self.classModel, data, pydanticModel)
                res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def update_one(
            self,
            pydanticModel,
            session: AsyncSession,
            whereClauseObject: Optional[BaseWhereClause] = None,
            **whereClause
        ):
        try:
            query = select(self.classModel)
            for k, attr in pydanticModel.dict().items():
                if k in vars(self.classModel):
                    query = query.where(vars(self.classModel)[k]==attr)
            if whereClauseObject:
                query = whereClauseObject.applyWhereObject(query, whereClause)
            data = await session.execute(query)
            data = data.scalars().first()
            if not data:
                res = create_response(status=status.data_is_not_updated())
            else:
                await update_data(
                    session, self.classModel, data,
                    pydanticModel.__dict__[self.classModel.__tablename__]
                    )
                res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def update_many(
            self,
            pydanticModelCollection,
            reference_key,
            session: AsyncSession
        ):
        try:
            dataCollection = pydanticModelCollection.__dict__[self.classModel.__tablename__]
            statuses = []
            successUpdate = 0
            failUpdate = 0
            for data in dataCollection:
                try:                    
                    query = select(self.classModel)
                    query = self.where().applyWhereObject(
                        query, {reference_key: data.dict()[reference_key]}
                        )
                    obj = await session.execute(query)
                    obj = obj.scalars().first()
                    if not obj:
                        failUpdate += 1
                        statuses.append(status.data_is_not_updated())
                    else:
                        await update_data(session, self.classModel, obj, data)
                        successUpdate += 1
                        statuses.append(status.success())
                except Exception as e:
                    logger.error(str(e))
                    failUpdate += 1
                    statuses.append(status.error(e))
            res = create_response(
                data={"status": statuses},
                meta={
                    "succeed": successUpdate,
                    "failed": failUpdate
                },
                status=status.success()
                )
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def delete(
            self,
            id: Optional[int],
            session: AsyncSession,
            whereClauseObject: Optional[BaseWhereClause] = None,
            **whereClause
        ):
        try:
            query = select(self.classModel)
            if id != None:
                query = query.where(self.classModel.id == id)
            if whereClauseObject:
                query = whereClauseObject.applyWhereObject(query, whereClause)
            data = await session.execute(query)
            data = data.scalars().all()
            if not data:
                res = create_response(status=status.data_is_not_exist())
            else:
                for d in data:
                    await session.delete(d)
                await session.commit()
                res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res
    
    async def delete_many(
            self,
            deleteParamsPydantic,
            session: AsyncSession
        ):
        try:
            statuses = []
            successDelete = 0
            failDelete = 0
            query = select(self.classModel)
            query = self.where().applyWhereObject(query, deleteParamsPydantic.dict())
            print(11111, query)
            datas = await session.execute(query)
            datas = datas.scalars().all()
            for data in datas:
                try:
                    if not data:
                        failDelete += 1
                        statuses.append(status.data_is_not_exist())
                    else:
                        await session.delete(data)
                        successDelete += 1
                        statuses.append(status.success())
                except Exception as e:
                    logger.error(e)
                    failDelete += 1
                    statuses.append(status.error(e))
            await session.commit()
            res = create_response(
                data={"status": statuses},
                meta={
                    "succeed": successDelete,
                    "failed": failDelete
                },
                status=status.success()
                )
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

def generate_pydantic_model(
        classModel: decl_api.DeclarativeMeta,
        modelName: str = "",
        exclude_attributes: Optional[List[Union[str, Column, InstrumentedAttribute]]] = [],
        include_attributes_default: Optional[Dict[str, Any]] = {},
        include_attributes_paramsType: Optional[
            Dict[str, Union[
                fastapi.Path, fastapi.Body, fastapi.Form, fastapi.Query, fastapi.Cookie, fastapi.Header
                ]]] = {},
        uniform_attributes_default: Optional[Any] = None,
        uniform_attributes_paramsType: Optional[Union[
            fastapi.Path, fastapi.Body, fastapi.Form, fastapi.Query, fastapi.Cookie, fastapi.Header
            ]] = None
    ):
    annots = get_annotation(classModel)
    for ex_at in exclude_attributes:
        if type(ex_at) != str:
            ex_at = ex_at.__str__().split(".")[1]
        if ex_at in annots:
            annots.pop(ex_at)
    for key, at_def in include_attributes_default.items():
        if key in annots:
            dataType, defVal = annots[key]
            annots[key] = (dataType, at_def)
    if uniform_attributes_paramsType:
        for key, an in annots.items():
            dataType, defVal = an
            if uniform_attributes_default:
                defVal = uniform_attributes_default
            try:
                if (
                        [
                            len(dataType.__args__) > 1,
                            type(None) in dataType.__args__,
                        ]
                    ):
                    annots[key] = (dataType, uniform_attributes_paramsType(default=defVal))
                else:
                    annots[key] = (dataType, uniform_attributes_paramsType(..., default=defVal))
            except:
                annots[key] = (dataType, uniform_attributes_paramsType(default=defVal))
    else:
        for key, at_ptype in include_attributes_paramsType.items():
            if key in annots:
                try:
                    dataType, defVal = annots[key]
                    if uniform_attributes_default:
                        defVal = uniform_attributes_default
                    if (
                            [
                                len(dataType.__args__) > 1,
                                type(None) in dataType.__args__,
                            ]
                        ):
                        if defVal:
                            annots[key] = (dataType, at_ptype(default=defVal))
                        else:
                            annots[key] = (dataType, at_ptype(...))
                    else:
                        annots[key] = (dataType, at_ptype(..., default=defVal))
                except:
                    annots[key] = (dataType, at_ptype(default=defVal))
    if not modelName: modelName = classModel.tablename+"Pydantic"
    ModelPydantic = create_model(modelName, **annots)
    return ModelPydantic

def get_annotation(classModel: decl_api.DeclarativeMeta):
    try:
        fields = [i for i in vars(classModel) if "_" not in [i[0], i[-1]]]
        keyValuePair = {}
        for f in fields:
            if "comparator" in classModel.__dict__[f].__dict__:
                if "column" in str(classModel.__dict__[f].__dict__["comparator"]).lower():
                    comp = classModel.__dict__[f].__dict__["comparator"]
                    if "VARCHAR" in str(comp.type):
                        if "enum_class" in comp.type.__dict__:
                            if comp.nullable:
                                defaultValue = comp.default.arg if comp.default else None
                                keyValuePair[f] = (Optional[comp.type.enum_class], defaultValue)
                            else:
                                defaultValue = comp.default.arg if comp.default else None
                                keyValuePair[f] = (comp.type.enum_class, defaultValue)
                        else:
                            if comp.nullable:
                                defaultValue = comp.default.arg if comp.default else None
                                keyValuePair[f] = (Optional[str], defaultValue)
                            else:
                                defaultValue = comp.default.arg if comp.default else None
                                keyValuePair[f] = (str, defaultValue)
                    elif "TEXT" in str(comp.type):
                        if comp.nullable:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (Optional[str], defaultValue)
                        else:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (str, defaultValue)
                    elif "BOOLEAN" in str(comp.type):
                        if comp.nullable:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (Optional[bool], defaultValue)
                        else:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (bool, defaultValue)
                    elif "INTEGER" in str(comp.type):
                        if comp.nullable:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (Optional[int], defaultValue)
                        else:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (int, defaultValue)
                    elif "FLOAT" in str(comp.type):
                        if comp.nullable:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (Optional[float], defaultValue)
                        else:
                            defaultValue = comp.default.arg if comp.default else None
                            keyValuePair[f] = (float, defaultValue)
                    elif "DATETIME" in str(comp.type):
                        if comp.nullable:
                            defaultValue = None
                            keyValuePair[f] = (Optional[datetime], defaultValue)
                        else:
                            defaultValue = None
                            keyValuePair[f] = (datetime, defaultValue)
        return keyValuePair
    except:
        return {}


import logging

logger = logging.getLogger()

# import logging, os, time, io
# from typing import List

# class CustomFilter(logging.Filter):
#     def __init__(self,logFormat,levels=[10,20,30,40,50],**kargs):
#         self.levels = levels
#         self._unapplied = ["_logFormat","_unapplied","filter","_applyAttr"]
#         self._logFormat = logFormat
#         if kargs:
#             for key in kargs:
#                 if not key in self.__dict__:
#                     setattr(self, key, kargs[key])

#     def _applyAttr(self, targetObject):
#         for attr, value in vars(self).items():
#             if (attr not in targetObject.__dict__) and (attr not in self._unapplied):
#                 setattr(targetObject, attr, value)
    
#     def _setLevels(self, *levels):
#         self.levels = levels

#     def filter(self, record):
#         self._applyAttr(record)
#         return record.levelno in self.levels

# class Streamer(logging.StreamHandler):
#     def __init__(self, filterObject):
#         super(Streamer, self).__init__()
#         self._formatter = logging.Formatter(filterObject._logFormat)
#         self.setFormatter(self._formatter)
#         self.addFilter(filterObject)

# class CustomLogger(logging.getLoggerClass()):
#     def __init__(self, name: str, disable: bool, filterObject: CustomFilter):
#         super(CustomLogger, self).__init__(name, 0)
#         self._streamer = Streamer(filterObject)
#         self.addHandler(self._streamer)
#         self.disabled = bool(disable)
#         self._filter = filterObject
    
#     def disable(self): 
#         self.disabled = True

#     def enable(self): 
#         self.disabled = False

#     def setLevels(self, levels: List[int]):
#         self._filter.levels = levels



# # globalFilter = CustomFilter(logFormat='[%(asctime)s][%(levelname)s] %(message)s', levels=Logger.levels)
# globalFilter = CustomFilter(
#     logFormat='[%(asctime)s][%(filename)s][%(levelname)s][line:%(lineno)s] - %(message)s'
#     )
# logger = CustomLogger(
#     name = "fastapi-simple-crud",
#     disable = False,
#     filterObject = globalFilter)

# """
# ## implementation examples
# # mock aja
# os.environ["SERVICE_NAME"] = "merchant-management"
# os.environ["LEVEL"] = "DEBUG"
# os.environ["LOG_DISABLE"] = "false"
# os.environ["ADMIN_SERVICE_NAME"] = "admin"
# os.environ["ADMIN_LEVEL"] = "DEBUG"
# os.environ["ADMIN_LOG_DISABLE"] = "false"


# # admin logger
# adminFilter = CustomFilter(
#         logFormat='%(asctime)s - %(name)s - %(filename)s - %(levelname)s line:%(lineno)s - %(user)s - %(status)s - %(message)s',
#         user="danang", status="mantap")

# adminLogger = CustomLogger(
#     name = os.getenv("ADMIN_SERVICE_NAME"),
#     level = os.getenv("ADMIN_LEVEL"),
#     disable = os.getenv("ADMIN_LOG_DISABLE") == "true",
#     filterObject = adminFilter)

# # general logger
# generalFilter = CustomFilter(
#         logFormat='%(asctime)s - %(name)s - %(filename)s - %(levelname)s line:%(lineno)s - %(user)s - %(status)s - %(message)s',
#         user="biasa", status="sad")

# logger = CustomLogger(
#     name = os.getenv("SERVICE_NAME"),
#     level = os.getenv("LEVEL"),
#     disable = os.getenv("LOG_DISABLE") == "true",
#     filterObject = generalFilter)

# adminLogger.info("aaaaaaaaa")
# """



from setuptools import setup, find_packages
import codecs
import os

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

VERSION = '0.1.3'
DESCRIPTION = 'Generate CRUD routers for you schema in a simple way'

# Setting up
setup(
    name="fastapi-simple-crud",
    version=VERSION,
    author="danangjoyoo (Agus Danangjoyo)",
    author_email="<agus.danangjoyo.blog@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=["uvicorn","fastapi","sqlalchemy","pydantic"],
    keywords=['fastapi', 'crud', 'restful', 'routing', "router", 'generator','aiosqlite'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Typing :: Typed"
    ]
)


from fastapi import FastAPI, APIRouter, Depends, Request, Path, Query, Body
from fastapi.encoders import SetIntStr, DictIntStrAny
from fastapi.routing import APIRoute
from fastapi.datastructures import Default
from pydantic import BaseConfig, BaseModel, create_model
from pydantic.fields import ModelField
from starlette.routing import BaseRoute
from starlette.responses import Response, JSONResponse
from starlette.types import ASGIApp
from typing import Optional, List, Sequence, Type, Any, Callable, Union, Dict
from sqlalchemy.orm import decl_api
from sqlalchemy.ext.asyncio import AsyncSession
from types import FunctionType
from datetime import datetime

from .dependencies.utils import BaseCRUD, generate_pydantic_model
from .dependencies.utility import CommonQueryGetter


class SimpleEndpoint:
    def __init__(self, 
            path: str = "",
            enable: bool = True,
            response_model: Optional[Type[Any]] = None,
            status_code: Optional[int] = None,
            tags: Optional[List[str]] = None,
            dependencies: Optional[Sequence[Depends]] = None,
            summary: Optional[str] = None,
            description: Optional[str] = None,
            response_description: str = "Successful Response",
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            deprecated: Optional[bool] = None,
            operation_id: Optional[str] = None,
            response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
            response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Type[Response] = Default(JSONResponse),
            name: Optional[str] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            openapi_extra: Optional[Dict[str, Any]] = None,
            pydantic_model: Optional[BaseModel] = None
        ):
        self.enable = enable
        self.path = path
        self.response_model = response_model
        self.status_code = status_code
        self.tags = tags
        self.dependencies = dependencies
        self.summary = summary
        self.description = description
        self.response_description = response_description
        self.responses = responses
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.response_model_include = response_model_include
        self.response_model_exclude = response_model_exclude
        self.response_model_by_alias = response_model_by_alias
        self.response_model_exclude_unset = response_model_exclude_unset
        self.response_model_exclude_defaults = response_model_exclude_defaults
        self.response_model_exclude_none = response_model_exclude_none
        self.include_in_schema = include_in_schema
        self.response_class = response_class
        self.name = name
        self.callbacks = callbacks
        self.openapi_extra = openapi_extra
        self.modelPydantic = pydantic_model
    
    def get_endpoint_kwargs(self, exclude_attributes: Optional[List[str]]=[]):
        params = vars(self).copy()
        for key in exclude_attributes:
            params.pop(key)
        return params    


class SimpleRouter(APIRouter):
    """
    Simple Router Class to define a router for SimpleCRUDGenerator
    """
    def __init__(
            self,
            classModel: decl_api.DeclarativeMeta,
            *,
            prefix: str = "",
            tags: Optional[List[str]] = None,
            dependencies: Optional[Sequence[Depends]] = None,
            default_response_class: Type[Response] = Default(JSONResponse),
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            routes: Optional[List[BaseRoute]] = None,
            redirect_slashes: bool = True,
            default: Optional[ASGIApp] = None,
            dependency_overrides_provider: Optional[Any] = None,
            route_class: Type[APIRoute] = APIRoute,
            on_startup: Optional[Sequence[Callable[[], Any]]] = None,
            on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
            deprecated: Optional[bool] = None,
            include_in_schema: bool = True,
            crud_create: Union[SimpleEndpoint, bool, None] = True,
            crud_read: Union[SimpleEndpoint, bool, None] = True,
            crud_update: Union[SimpleEndpoint, bool, None] = True,
            crud_delete: Union[SimpleEndpoint, bool, None] = True,
            disable_crud: bool = False
        ):
        self.classModel = classModel
        self.tablename = classModel.__tablename__
        self.modelPydanticforCreate = generate_pydantic_model(self.classModel, modelName=self.tablename+"PydanticSimpleCreate")
        self.modelPydanticforUpdate = generate_pydantic_model(self.classModel, modelName=self.tablename+"PydanticSimpleUpdate", exclude_attributes=["id"])
        self.crud = BaseCRUD(classModel)
        if not tags: tags = [self.tablename]
        if not prefix: prefix = f"/{self.tablename}"
        super().__init__(
                prefix=prefix,
                tags=tags,
                dependencies=dependencies,
                default_response_class=default_response_class,
                responses=responses,
                callbacks=callbacks,
                routes=routes,
                redirect_slashes=redirect_slashes,
                default=default,
                dependency_overrides_provider=dependency_overrides_provider,
                route_class=route_class,
                on_startup=on_startup,
                on_shutdown=on_shutdown,
                deprecated=deprecated,
                include_in_schema=include_in_schema)
        if disable_crud:
            crud_create = None
            crud_read = None
            crud_update = None
            crud_delete = None
        if type(crud_create) == SimpleEndpoint:
            self.crud_create = crud_create
        else:
            if crud_create:
                self.crud_create = SimpleEndpoint(enable=True)
            else:
                self.crud_create = SimpleEndpoint(enable=False)
        if type(crud_read) == SimpleEndpoint:
            self.crud_read = crud_read
        else:
            if crud_read:
                self.crud_read = SimpleEndpoint(enable=True)
            else:
                self.crud_read = SimpleEndpoint(enable=False)
        if type(crud_update) == SimpleEndpoint:
            self.crud_update = crud_update
        else:
            if crud_update:
                self.crud_update = SimpleEndpoint(path="/{id}", enable=True)
            else:
                self.crud_update = SimpleEndpoint(enable=False)
        if type(crud_delete) == SimpleEndpoint:
            self.crud_delete = crud_delete
        else:
            if crud_delete:
                self.crud_delete = SimpleEndpoint(path="/{id}", enable=True)
            else:
                self.crud_delete = SimpleEndpoint(enable=False)
        self._get_session = None
    
    def set_the_get_session(self, method: FunctionType):
        self._get_session = method
        
    def _setup_crud(self):
        if self.crud_create.enable:
            kargs = self.crud_create.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "create "+self.tablename
            if self.crud_create.modelPydantic:
                modelPydantic_ = self.crud_create.modelPydantic
            else:
                modelPydantic_ = self.modelPydanticforCreate
            @self.post(**kargs)
            async def base_post(
                    request: Request,
                    modelPydantic: modelPydantic_,
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.create(modelPydantic, session)
        
        if self.crud_read.enable:
            # kargs = self.crud_read.get_endpoint_kwargs(
            #     exclude_attributes=["enable","modelPydantic"]
            #     )
            # if not kargs["name"]:
            #     kargs["name"] = "read "+self.tablename
            # @self.get(**kargs)
            # async def base_get(
            #         request: Request,
            #         getParams = Depends(CommonQueryGetter),
            #         session: AsyncSession = Depends(self._get_session)
            #     ):
            #     return await self.crud.read(getParams, session)
            kargs = self.crud_read.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "read many "+self.tablename
            if self.crud_read.modelPydantic:
                modelPydantic_ = self.crud_read.modelPydantic
            else:
                modelPydantic_ = generate_pydantic_model(
                    classModel=self.classModel,
                    modelName=self.tablename+"PydanticSimpleReadMany",
                    uniform_attributes_paramsType=Query
                )
            @self.get(**kargs)
            async def base_get_many(
                    request: Request,
                    readParams = Depends(modelPydantic_),
                    getParams = Depends(CommonQueryGetter),
                    session: AsyncSession = Depends(self._get_session)
                ):
                wc = self.crud.where(**readParams.dict())
                return await self.crud.read_many(getParams, session, wc)

        if self.crud_update.enable:
            # kargs = self.crud_update.get_endpoint_kwargs(
            #     exclude_attributes=["enable","modelPydantic"]
            #     )
            # if not kargs["name"]:
            #     kargs["name"] = "update "+self.tablename
            # if self.crud_update.modelPydantic:
            #     modelPydantic_ = self.crud_update.modelPydantic
            # else:
            #     modelPydantic_ = self.modelPydanticforUpdate
            # @self.put(**kargs)
            # async def base_put(
            #         request: Request,
            #         modelPydantic: modelPydantic_,
            #         id: int = Path(...,min=1),
            #         session: AsyncSession = Depends(self._get_session)
            #     ):
            #     return await self.crud.update(modelPydantic, id, session)
            kargs = self.crud_update.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "update one "+self.tablename
            if self.crud_update.modelPydantic:
                modelPydantic_ = self.crud_update.modelPydantic
            else:
                premodelPydantic_ = generate_pydantic_model(
                    classModel=self.classModel,
                    modelName=self.tablename+"PydanticSimpleUpdateOne",
                    exclude_attributes=["id"],
                )
                modelPydantic_ = create_model(
                    self.tablename+"PydanticSimpleUpdateOnePacked",
                    **{
                        "id": (int, Path(...)),
                        self.tablename: (premodelPydantic_, Body(...))
                    }
                )
            @self.put(**kargs)
            async def base_put(
                    request: Request,
                    modelPydantic: modelPydantic_ = Depends(),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.update_one(modelPydantic, session)

        if self.crud_delete.enable:
            kargs = self.crud_delete.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "delete "+self.tablename
            @self.delete(**kargs)
            async def base_delete(
                    request: Request,
                    id: int = Path(...,min=1),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.delete(id, session)


class ExtendedRouter(SimpleRouter):
    """
    Extended Version of Simple Router.

    Provide the following API:
    - create one
    - create many
    - read one
    - read many
    - update one
    - update many
    - delete one
    - delete many
    """
    def __init__(
            self,
            classModel: decl_api.DeclarativeMeta,
            *,
            prefix: str = "",
            tags: Optional[List[str]] = None,
            dependencies: Optional[Sequence[Depends]] = None,
            default_response_class: Type[Response] = Default(JSONResponse),
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            routes: Optional[List[BaseRoute]] = None,
            redirect_slashes: bool = True,
            default: Optional[ASGIApp] = None,
            dependency_overrides_provider: Optional[Any] = None,
            route_class: Type[APIRoute] = APIRoute,
            on_startup: Optional[Sequence[Callable[[], Any]]] = None,
            on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
            deprecated: Optional[bool] = None,
            include_in_schema: bool = True,
            create_one: Union[SimpleEndpoint, bool, None] = True,
            create_many: Union[SimpleEndpoint, bool, None] = True,
            read_one: Union[SimpleEndpoint, bool, None] = True,
            read_many: Union[SimpleEndpoint, bool, None] = True,
            # read_many_like: Union[SimpleEndpoint, bool, None] = True,
            update_one: Union[SimpleEndpoint, bool, None] = True,
            update_many: Union[SimpleEndpoint, bool, None] = True,
            delete_one: Union[SimpleEndpoint, bool, None] = True,
            delete_many: Union[SimpleEndpoint, bool, None] = True,
            disable_crud: bool = False
        ):
        super().__init__(
                classModel=classModel,
                prefix=prefix,
                tags=tags,
                dependencies=dependencies,
                default_response_class=default_response_class,
                responses=responses,
                callbacks=callbacks,
                routes=routes,
                redirect_slashes=redirect_slashes,
                default=default,
                dependency_overrides_provider=dependency_overrides_provider,
                route_class=route_class,
                on_startup=on_startup,
                on_shutdown=on_shutdown,
                deprecated=deprecated,
                include_in_schema=include_in_schema,
                disable_crud=True)
   
        if disable_crud:
            create_one = None
            create_many = None
            read_one = None
            read_many = None
            update_one = None
            update_many = None
            delete_one = None
            delete_many = None

        # create one
        if type(create_one) == SimpleEndpoint:
            self.create_one = create_one
        else:
            if create_one:
                self.create_one = SimpleEndpoint(path="/one",enable=True)
            else:
                self.create_one = SimpleEndpoint(enable=False)

        # create many
        if type(create_many) == SimpleEndpoint:
            self.create_many = create_many
        else:
            if create_many:
                self.create_many = SimpleEndpoint(path="",enable=True)
            else:
                self.create_many = SimpleEndpoint(enable=False)

        # read one
        if type(read_one) == SimpleEndpoint:
            self.read_one = read_one
        else:
            if read_one:
                self.read_one = SimpleEndpoint(path="/one", enable=True)
            else:
                self.read_one = SimpleEndpoint(enable=False)
        
        # read many
        if type(read_many) == SimpleEndpoint:
            self.read_many = read_many
        else:
            if read_many:
                self.read_many = SimpleEndpoint(path="", enable=True)
            else:
                self.read_many = SimpleEndpoint(enable=False)
        
        # # read many like
        # if type(read_paginate) == SimpleEndpoint:
        #     self.read_paginate = read_paginate
        # else:
        #     if read_paginate:
        #         self.read_paginate = SimpleEndpoint(path="/like", enable=True)
        #     else:
        #         self.read_paginate = SimpleEndpoint(enable=False)
        
        # update one
        if type(update_one) == SimpleEndpoint:
            self.update_one = update_one
        else:
            if update_one:
                self.update_one = SimpleEndpoint(path="/{id}", enable=True)
            else:
                self.update_one = SimpleEndpoint(enable=False)
        
        # update many
        if type(update_many) == SimpleEndpoint:
            self.update_many = update_many
        else:
            if update_many:
                self.update_many = SimpleEndpoint(path="", enable=True)
            else:
                self.update_many = SimpleEndpoint(enable=False)

        # delete one
        if type(delete_one) == SimpleEndpoint:
            self.delete_one = delete_one
        else:
            if delete_one:
                self.delete_one = SimpleEndpoint(path="/{id}", enable=True)
            else:
                self.delete_one = SimpleEndpoint(enable=False)
        
        # delete many
        if type(delete_many) == SimpleEndpoint:
            self.delete_many = delete_many
        else:
            if delete_many:
                self.delete_many = SimpleEndpoint(path="", enable=True)
            else:
                self.delete_many = SimpleEndpoint(enable=False)
    
    def _setup_crud(self):
        if self.create_one.enable:
            kargs = self.create_one.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "create one "+self.tablename
            if self.create_one.modelPydantic:
                modelPydantic_ = self.create_one.modelPydantic
            else:
                modelPydantic_ = generate_pydantic_model(
                    classModel=self.classModel,
                    modelName=self.tablename+"PydanticSimpleCreateOne"
                )
            @self.post(**kargs)
            async def base_post_one(
                    request: Request,
                    modelPydantic: modelPydantic_,
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.create(modelPydantic, session)

        if self.create_many.enable:
            kargs = self.create_many.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "create many "+self.tablename
            if self.create_many.modelPydantic:
                modelPydantic_ = self.create_many.modelPydantic
            else:
                modelPydantic_ = create_model(
                    self.tablename+"PydanticSimpleCreateMany",
                    **{self.tablename: (
                        Optional[
                            List[
                                generate_pydantic_model(
                                    classModel=self.classModel,
                                    modelName=self.tablename+"PydanticSimpleCreateOneForMany"
                                    )
                                ]
                            ],
                        None
                        )
                    }
                )
            @self.post(**kargs)
            async def base_post_many(
                    request: Request,
                    modelPydantic: modelPydantic_,
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.create_many(modelPydantic, session)

        if self.read_one.enable:
            kargs = self.read_one.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "read one "+self.tablename
            if self.read_one.modelPydantic:
                modelPydantic_ = self.read_one.modelPydantic
            else:
                modelPydantic_ = create_model(
                    self.tablename+"PydanticSimpleReadOne", id=(int, Query(...))
                    )
            @self.get(**kargs)
            async def base_get_one(
                    request: Request,
                    modelPydantic_ = Depends(modelPydantic_),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.read_one(modelPydantic_, session)
        
        if self.read_many.enable:
            kargs = self.read_many.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "read many "+self.tablename
            if self.read_many.modelPydantic:
                modelPydantic_ = self.read_many.modelPydantic
            else:
                modelPydantic_ = generate_pydantic_model(
                    classModel=self.classModel,
                    modelName=self.tablename+"PydanticSimpleReadMany",
                    uniform_attributes_paramsType=Query
                )
            @self.get(**kargs)
            async def base_get_many(
                    request: Request,
                    readParams = Depends(modelPydantic_),
                    getParams = Depends(CommonQueryGetter),
                    session: AsyncSession = Depends(self._get_session)
                ):
                wc = self.crud.where(**readParams.dict())
                return await self.crud.read_many(getParams, session, wc)

        # if self.read_paginate.enable:
        #     kargs = self.read_paginate.get_endpoint_kwargs(
        #         exclude_attributes=["enable","modelPydantic"]
        #         )
        #     if not kargs["name"]:
        #         kargs["name"] = "read paginate "+self.tablename
        #     @self.get(**kargs)
        #     async def base_get_paginate(
        #             request: Request,
        #             getParams = Depends(CommonQueryGetter),
        #             session: AsyncSession = Depends(self._get_session)
        #         ):
        #         return await self.crud.read(getParams, session)

        if self.update_one.enable:
            kargs = self.update_one.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "update one "+self.tablename
            if self.update_one.modelPydantic:
                modelPydantic_ = self.update_one.modelPydantic
            else:
                premodelPydantic_ = generate_pydantic_model(
                    classModel=self.classModel,
                    modelName=self.tablename+"PydanticSimpleUpdateOne",
                    exclude_attributes=["id"],
                )
                modelPydantic_ = create_model(
                    self.tablename+"PydanticSimpleUpdateOnePacked",
                    **{
                        "id": (int, Path(...)),
                        self.tablename: (premodelPydantic_, Body(...))
                    }
                )
            @self.put(**kargs)
            async def base_put_one(
                    request: Request,
                    modelPydantic: modelPydantic_ = Depends(),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.update_one(modelPydantic, session)
        
        if self.update_many.enable:
            kargs = self.update_many.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "update many "+self.tablename
            if self.update_many.modelPydantic:
                modelPydantic_ = self.update_many.modelPydantic
            else:
                modelPydantic_ = create_model(
                    self.tablename+"PydanticSimpleUpdateMany",
                    **{self.tablename: (Optional[List[
                        generate_pydantic_model(
                            classModel=self.classModel,
                            modelName=self.tablename+"PydanticSimpleUpdateOneWithID"
                        )]], None)}
                )
            @self.put(**kargs)
            async def base_put_many(
                    request: Request,
                    pydanticModelCollection: modelPydantic_,
                    reference_key: str = Query(
                        ...,
                        description="Put your reference key that will be used to refer your data and won't be updated"
                        ),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.update_many(reference_key, pydanticModelCollection, session)

        if self.delete_one.enable:
            kargs = self.delete_one.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "delete one "+self.tablename
            @self.delete(**kargs)
            async def base_delete_one(
                    request: Request,
                    id: int = Path(...,min=1),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.delete(id, session)
        
        if self.delete_many.enable:
            kargs = self.delete_many.get_endpoint_kwargs(
                exclude_attributes=["enable","modelPydantic"]
                )
            if not kargs["name"]:
                kargs["name"] = "delete many "+self.tablename
            if self.delete_many.modelPydantic:
                modelPydantic_ = self.delete_many.modelPydantic
            else:
                modelPydantic_ = generate_pydantic_model(
                    classModel=self.classModel,
                    modelName=self.tablename+"PydanticSimpleDeleteMany",
                    uniform_attributes_paramsType=Query
                )
            @self.delete(**kargs)
            async def base_delete_many(
                    request: Request,
                    deleteParams = Depends(modelPydantic_),
                    session: AsyncSession = Depends(self._get_session)
                ):
                return await self.crud.delete_many(deleteParams, session)


RouterClasses = [SimpleRouter, ExtendedRouter]
SimpleRouterType = Union[SimpleRouter, ExtendedRouter]